/*
 * Copyright (C) 2004 Pedro Luis Lucas Rosado
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA. 
 */



#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include <string.h> /* For strlen */

#include <glib/gutils.h>

#include <libgnome/gnome-i18n.h>

#include <gedit/gedit-menus.h>
#include <gedit/gedit-plugin.h>
#include <gedit/gedit-debug.h>
#include <gedit/gedit-prefs-manager.h>

#define MENU_ITEM_LABEL		N_("Autocomplete")
#define MENU_ITEM_PATH		"/menu/Tools/ToolsOps_4/"
#define MENU_ITEM_NAME		"Autocomplete"
#define MENU_ITEM_TIP		N_("Autocomplete words")

G_MODULE_EXPORT GeditPluginState update_ui (GeditPlugin *plugin, BonoboWindow *window);
G_MODULE_EXPORT GeditPluginState destroy (GeditPlugin *pd);
G_MODULE_EXPORT GeditPluginState activate (GeditPlugin *pd);
G_MODULE_EXPORT GeditPluginState deactivate (GeditPlugin *pd);
G_MODULE_EXPORT GeditPluginState init (GeditPlugin *pd);


/** This is the struct representing plugin data.*/
struct _Autocomplete {
	gboolean active;	/*Is plugin active*/
	gint jump; 		/*Place to insert word. If jump<0 --> No insert. */
	GString *word; 		/*Autocomplete word*/
	gchar last_char;	/*Last char typed*/
	gint conection_id; 	/*Is the id of signal handler to text_view*/
	gint conection_id_key_press; 	/*Is the id of signal handler to text_view*/
	GtkTextView *text_view;	/*Actual text view*/
	GCompletion *completion;/*GLib autocompletion tool*/
	GList *wl;		/*List of words in document*/
	gboolean key_press;
} ;

/* Definions */
typedef struct _Autocomplete Autocomplete;

/** This def returns the struct representing plugin data. */
#define get_autocomplete(pd) ((Autocomplete *)((pd)->private_data))

static void update_word_list(GeditPlugin *pd);

/** This function compares two similar strings. It returns the diferent
 * part of strings. w1 >= w2. Ex.:
 * gchar *a = "123456";
 * gchar *b = "1234";
 * gchar *c = word_substraction( a , b ); // c="56";
 */
static gchar *
word_substraction(gchar *w1, gchar *w2)
{

	while(
		*w2!='\0' && *w1!='\0' &&
		( g_utf8_get_char (w1)==g_utf8_get_char (w2) )
	     )
	{
		w1 = g_utf8_next_char(w1);
		w2 = g_utf8_next_char(w2);
	}

	return w1;
}

static gboolean
key_press_event_callback(GtkWidget *widget,
			GdkEventKey *event,
			gpointer user_data)
{
	Autocomplete *ac = (Autocomplete *)get_autocomplete(  (GeditPlugin *)user_data  );

	if(g_unichar_isgraph(gdk_keyval_to_unicode (event->keyval)))
		ac->key_press = TRUE;
	else
		ac->key_press = FALSE;
	return FALSE;
}

/** If key is pressed , this function is called. If it's necessary, word is
 *  autocomplete.
 */
static gint
autocomplete_callback(GtkTextBuffer *textbuffer,
                                            GtkTextIter *arg1,
                                            gchar *arg2,
                                            gint arg3,
                                            gpointer user_data)
{
	Autocomplete *ac = (Autocomplete *)get_autocomplete(  (GeditPlugin *)user_data  );
	GtkTextBuffer /*GeditDocument*/ *doc;
	GtkTextIter iter;
	GtkTextIter start_iter;
	gint cursor;

	/* Checks if plugin is active */
	if( !ac->key_press || ac->active==FALSE ) return TRUE;

	/* Checks if readable key has been pressed. */
	if( g_utf8_strlen(arg2,-1) > 1 ) return TRUE;

	ac->last_char = *arg2;

	doc = GTK_TEXT_BUFFER ( gedit_get_active_document () );

	gtk_text_buffer_begin_user_action (GTK_TEXT_BUFFER (doc));

	/* Finds cursor position in text. */
	gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (doc),
                                    &iter,
                                    gtk_text_buffer_get_mark (GTK_TEXT_BUFFER (doc),
					                      "insert"));

	cursor = gtk_text_iter_get_offset (&iter);

	/* Auto indent could causes plugin malfunction. It must be controlled. */
	if( gedit_prefs_manager_get_auto_indent() )
		gedit_view_set_auto_indent( GEDIT_VIEW ( gedit_get_active_view() ), TRUE);

	/* If user press enter and word has been autocompleted, word is accepted. */
	if( *arg2=='\n' && ac->jump>=0 && ac->jump==(cursor-1) )
	{
		start_iter=iter;
		gtk_text_iter_backward_char( &start_iter );

		gtk_text_buffer_delete (GTK_TEXT_BUFFER (doc), &start_iter, &iter);

		gtk_text_buffer_insert (GTK_TEXT_BUFFER (doc), &iter, ac->word->str, -1);

		/*I need stop emmission of "insert-text". Sometimes Spell plugin can generate errors.*/
		g_signal_stop_emission_by_name (doc, "insert-text");

		gtk_text_buffer_end_user_action (GTK_TEXT_BUFFER (doc));

		ac->jump = -1;

		return TRUE;
	}

	if( *arg2==' ' || *arg2=='\t' )
	{
		ac->jump=-1;

		update_word_list((GeditPlugin *)user_data);
	}


	/* Maybe word has been inserted by user. It must autocomplete word or update word list*/
	if(  cursor!=0  )
	{
		/* Finds word start. */
		start_iter=iter;

		if( gtk_text_iter_backward_word_start (&start_iter) ) {
			gchar *p, *text;

			text = gtk_text_iter_get_slice ( &start_iter, &iter );

			g_completion_complete(ac->completion,text,&p);

			if( p!=NULL ){
				gchar *r = word_substraction(p,text);

				if(*r!='\0'){
					gtk_text_buffer_insert (GTK_TEXT_BUFFER (doc), &iter, r, -1);

					ac->jump = cursor;

					start_iter=iter;

					gtk_text_iter_set_offset ( &start_iter, cursor);

					gtk_text_buffer_place_cursor (GTK_TEXT_BUFFER (doc), &iter);

					gtk_text_buffer_move_mark_by_name (GTK_TEXT_BUFFER (doc),
						"selection_bound", &start_iter);

					/*I need stop emmission of "insert-text". Sometimes Spell plugin can 
					 *generate errors.*/
					g_signal_stop_emission_by_name (doc, "insert-text");

					g_string_assign(ac->word,r);

					/* Auto indent causes plugin malfunction. Deactivates it in */
					/* autocomplete insertion. */
					if(gedit_prefs_manager_get_auto_indent())
						gedit_view_set_auto_indent(
							GEDIT_VIEW (gedit_get_active_view()) ,
							FALSE
						);

				}

				g_free(p);
				g_free(text);

			} 
			else 
			{
				/*
				ac->jump=-1;

				update_word_list((GeditPlugin *)user_data);
				*/
			}

		} 
		else 
		{ /* Update word list. */
			update_word_list((GeditPlugin *)user_data);
				
			ac->jump=-1;
		}
	
	}

	gtk_text_buffer_end_user_action (GTK_TEXT_BUFFER (doc));

	return TRUE;
}

/* You can use or hash table or double linked list to generate word list. */
/* Comment the next line to use double linked list. */
#define USE_HASH_TABLE
#ifdef USE_HASH_TABLE
static void
create_list_from_hash(	gpointer key,
			gpointer value,
			gpointer user_data)
{
	GList *wl= *( (GList **) user_data);

	wl = g_list_append( wl, g_strdup( (gchar*)key ) );

	*( (GList **) user_data) =wl;
}
#endif

/** Updates word list of document
 */
static void
update_word_list(GeditPlugin *pd)
{
	GeditDocument *doc;
	
	if( get_autocomplete(pd)->active==FALSE ) return;
	
	doc = gedit_get_active_document ();
	
	if (!( (doc == NULL) || (gedit_document_is_readonly (doc)) ) ) {
		GList *wl=NULL; /*Word list*/
		gchar *text=NULL;
#ifdef USE_HASH_TABLE
		GHashTable *hash_table = g_hash_table_new_full  ( g_str_hash,  g_str_equal, g_free, g_free );
#else
		int number_of_words=0;
#endif

		GtkTextIter iter_end, start_iter;

		gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (doc),
                                    &iter_end,
                                    gtk_text_buffer_get_mark (GTK_TEXT_BUFFER (doc),
					                      "insert"));

		if ( !gtk_text_iter_backward_word_starts ( &iter_end, 1000 ) )
                        gtk_text_buffer_get_start_iter( GTK_TEXT_BUFFER(doc), &iter_end);

		/* Finds words in text. */
		/*Keeps 2000 words in word list.*/
		while
			(
				gtk_text_iter_forward_word_end(&iter_end)
					&&
#ifdef USE_HASH_TABLE
				g_hash_table_size (hash_table) <= 2000
#else
				number_of_words <= 2000
#endif
			)
		{
			start_iter=iter_end;
			gtk_text_iter_backward_word_start (&start_iter);

			text=gtk_text_iter_get_slice(&start_iter,&iter_end);

#ifdef USE_HASH_TABLE
			if ( strlen(text)>2 )
				g_hash_table_insert( hash_table, text, NULL );
			else
				g_free(text);
#else
			if
			(
				strlen(text)>2
					&&
				NULL == g_list_find_custom ( wl, (gconstpointer) text, (GCompareFunc)strcmp )
			)
			{
				wl=g_list_insert_sorted(wl, (gpointer) text, (GCompareFunc)strcmp );
				number_of_words++;
			}
			else
				g_free(text);
#endif
		}

#ifdef USE_HASH_TABLE
		g_hash_table_foreach ( hash_table, create_list_from_hash, (gpointer) &wl);

		g_hash_table_destroy ( hash_table );
#endif

			/*
				GList *l=g_list_first(wl);
				printf("\n");
				int count=0;
				while( l!=NULL  )
				{
					printf("Palabra: %d %s\n", ++count, (gchar*)l->data);
					l=l->next;
				}
				//printf("%d\n",number_list_elements);
			*/



		/* Prepares g_completion GLib util. */
		g_completion_clear_items (get_autocomplete(pd)->completion);

		wl=g_list_first(wl);

		if(g_list_length(wl)!=0) {
			g_completion_add_items (get_autocomplete(pd)->completion, wl );

			/* Frees previus word list. */
			{
				GList *l=g_list_first( get_autocomplete(pd)->wl );

				while (l!=NULL) {
					g_free(l->data);

					l=l->next;
				}

				g_list_free(get_autocomplete(pd)->wl);
			}

			get_autocomplete(pd)->wl=wl;

		}

	}
	
}



/** This function is part of gedit plugin interface. When active document is
 *  changed, this function is called. Updates word list of document and
 *  connect "insert-text" signals.
 */
G_MODULE_EXPORT GeditPluginState
update_ui (GeditPlugin *pd, BonoboWindow *window)
{
	/*BonoboUIComponent *uic;*/

	GeditDocument *doc;

	gedit_debug (DEBUG_PLUGINS, "");

	if(get_autocomplete(pd)->text_view!=NULL){
		/*I must disconnect signals*/
		g_signal_handler_disconnect(
			G_OBJECT(gtk_text_view_get_buffer(get_autocomplete(pd)->text_view)),
			get_autocomplete(pd)->conection_id);

		g_signal_handler_disconnect(
			G_OBJECT(get_autocomplete(pd)->text_view),
			get_autocomplete(pd)->conection_id_key_press);

		get_autocomplete(pd)->text_view=NULL;
	}

	/*if(get_autocomplete(pd)->active==FALSE) return PLUGIN_OK;*/
	
	doc = gedit_get_active_document ();
	
	if (!( (doc == NULL) || (gedit_document_is_readonly (doc)) ) ) {
		/*I must connect signals*/
		GeditView* gedit_view=GEDIT_VIEW( gedit_get_active_view() );
		
		get_autocomplete(pd)->text_view=gedit_view_get_gtk_text_view(gedit_view);

		if ( get_autocomplete(pd)->text_view==NULL ) return PLUGIN_OK;

		get_autocomplete(pd)->conection_id_key_press=
			g_signal_connect_after(
				G_OBJECT(get_autocomplete(pd)->text_view),
				"key-release-event",
				G_CALLBACK(key_press_event_callback),
				(gpointer)(pd)
			);

		get_autocomplete(pd)->conection_id=
			g_signal_connect_after(G_OBJECT(
				gtk_text_view_get_buffer(get_autocomplete(pd)->text_view)),
				"insert-text",
				G_CALLBACK(autocomplete_callback),
				(gpointer)(pd)
			);

		gtk_source_buffer_set_check_brackets (GTK_SOURCE_BUFFER(doc), TRUE);
		{
			static GtkSourceTagStyle *style=NULL;
			if( style == NULL ){
				style= gtk_source_tag_style_new ();
				style->underline=TRUE;
				/*style->bold=TRUE;*/
			}
			gtk_source_buffer_set_bracket_match_style (
				GTK_SOURCE_BUFFER(doc), style);
		}
	}

	get_autocomplete(pd)->jump=-1;

	update_word_list(pd);


	/*
	g_return_val_if_fail (window != NULL, PLUGIN_ERROR);

	uic = gedit_get_ui_component_from_window (window);

	doc = gedit_get_active_document ();

	if ((doc == NULL) || (gedit_document_is_readonly (doc)))
		gedit_menus_set_verb_sensitive (uic, "/commands/" MENU_ITEM_NAME, FALSE);
	else
		gedit_menus_set_verb_sensitive (uic, "/commands/" MENU_ITEM_NAME, TRUE);

	printf("Llamada a update_ui nueva\n");
	*/

	return PLUGIN_OK;
	
}


/** This function is part of gedit plugin interface. It destroies plugin.
 */
G_MODULE_EXPORT GeditPluginState
destroy (GeditPlugin *pd)
{
	gedit_debug (DEBUG_PLUGINS, "");
	
	if(get_autocomplete(pd)->text_view!=NULL){
	
		/*I must disconnect the signals*/
		g_signal_handler_disconnect(
			G_OBJECT(gtk_text_view_get_buffer(get_autocomplete(pd)->text_view)),
			get_autocomplete(pd)->conection_id);
			
		get_autocomplete(pd)->text_view=NULL;
		
	}
	
	/* Frees word list. */
	{
		GList *l=g_list_first( get_autocomplete(pd)->wl );
		
		while (l!=NULL) {
			g_free(l->data);
			l=l->next;
		}
		
		g_list_free(get_autocomplete(pd)->wl);
		
	}

	g_string_free( get_autocomplete(pd)->word, TRUE );

	g_completion_free (get_autocomplete(pd)->completion);

	g_free(get_autocomplete(pd));

	return PLUGIN_OK;
}


/** This function deactivates plugin when menu_item_toggle is active.
 */
static void
activate_plugin_callback (BonoboUIComponent           *ui_component,
	     const char                  *path,
	     Bonobo_UIComponent_EventType type,
	     const char                  *state,
	     gpointer               	  data)
{
	gboolean s;

	gedit_debug (DEBUG_PLUGINS, "%s toggled to '%s'", path, state);

	s = (strcmp (state, "1") != 0);

	if (!s)
	{
		get_autocomplete( ((GeditPlugin *) data) ) -> active = TRUE;
		update_word_list( (GeditPlugin *) data );
	}
	else
		get_autocomplete( ((GeditPlugin *) data) ) -> active = FALSE;
}


/** This function is part of gedit plugin interface. It activates plugin.
 */
G_MODULE_EXPORT GeditPluginState
activate (GeditPlugin *pd)
{
	GList *top_windows;
        gedit_debug (DEBUG_PLUGINS, "");

        get_autocomplete(pd)->active=TRUE;

        top_windows = gedit_get_top_windows ();
        g_return_val_if_fail (top_windows != NULL, PLUGIN_ERROR);

        while (top_windows)
        {
		gedit_menus_add_menu_item_toggle (BONOBO_WINDOW (top_windows->data),
				MENU_ITEM_PATH, MENU_ITEM_NAME,
				MENU_ITEM_LABEL, MENU_ITEM_TIP,
				activate_plugin_callback, (gpointer) pd
				);

		gedit_menus_set_verb_state (
				gedit_get_ui_component_from_window (
					BONOBO_WINDOW (top_windows->data)
					),
                                "/commands/" MENU_ITEM_NAME,
				TRUE
				);

                pd->update_ui (pd, BONOBO_WINDOW (top_windows->data));

                top_windows = g_list_next (top_windows);
        }

        return PLUGIN_OK;
}



/** This function is part of gedit plugin interface. It deactivates plugin.
 */
G_MODULE_EXPORT GeditPluginState
deactivate (GeditPlugin *pd)
{
	gedit_menus_remove_menu_item_all (MENU_ITEM_PATH, MENU_ITEM_NAME);
	
	get_autocomplete(pd)->active=FALSE;

	return PLUGIN_OK;
}


/** This function is part of gedit plugin interface. It inits plugin.
 */
G_MODULE_EXPORT GeditPluginState
init (GeditPlugin *pd)
{

	/* initialize */
	Autocomplete *ac=g_new(Autocomplete, 1);

	gedit_debug (DEBUG_PLUGINS, "");

	pd->private_data = ac;

	ac->active	= FALSE;
	ac->text_view	= NULL;
	ac->completion	= g_completion_new (NULL);
	ac->wl		= NULL;
	ac->jump	= -1;
	ac->last_char	= ' ';
	ac->word	= g_string_new ("");
	ac->key_press   = FALSE;

	return PLUGIN_OK;

}





